// =================================================================================================
// -------------------------------------------------------------------------------------------------
// SSF Compactor
// -------------------------------------------------------------------------------------------------

#include <stdio.h>
#include "Memory.h"

// =================================================================================================
// -------------------------------------------------------------------------------------------------
// Subroutine to read a value from the list
// -------------------------------------------------------------------------------------------------

int ReadValue (FILE *List)

{
	int i0, i1, i2, i3, i4, i5, i6, i7;
	i0 = 0x00, i1 = 0x00;
	while (i0 < '0' || i0 > '9' && i0 < 'A' || i0 > 'F')
	{
		i0 = fgetc (List) & 0xFF;
		if (i0 == ';')
		{
			do
			{
				i0 = fgetc (List) & 0xFF;
			}
			while (i0 != 0x0A);
		}
	}
	do
	{
		if (i0 > '9')
		{
			i0 -= 'A' - ('9' + 0x01);
		}
		i0 -= '0';
		i1 <<= 0x04;
		i1 |= i0;
		i0 = fgetc (List) & 0xFF;
	}
	while (i0 >= '0' && i0 <= '9' || i0 >= 'A' && i0 <= 'F');
	return (i1);
}

// =================================================================================================
// -------------------------------------------------------------------------------------------------
// Subroutine to pad out a seciton
// -------------------------------------------------------------------------------------------------

void PadData (int Start, int End, char *&Memory, int PadMode)

{
	if (PadMode == 0x00)
	{
		while (Start < End)
		{
			Memory [Start++] = 0x00;
		}
	}
	else
	{
		int i0, i1 = 0x00;
		char PadName [] = "PAD128";
		while (Start < End)
		{
			i0 = PadName [i1++] & 0xFF;
			if (i0 == 0x00)
			{
				i1 = 0x00;
				i0 = PadName [i1++] & 0xFF;
			}
			Memory [Start++] = i0;
		}
	}
}

// =================================================================================================
// -------------------------------------------------------------------------------------------------
// Subroutine to compact a folder full of files into a .ssf ( CES) file
// -------------------------------------------------------------------------------------------------

void CompactFile (char *Directory)

{
	int DirectoryLoc, i0, i1, i2, i3, i4, i5, i6, i7;
	int Fail = 0x00, FileCount, StringSize, StringPos, FileNameLoc, NameType, PadSize, PadMode, Type, Offset, Size;
	for (DirectoryLoc = -0x01, i0 = ' '; i0 != 0x00; )
	{
		i0 = Directory [++DirectoryLoc] & 0xFF;
	}
	Directory [DirectoryLoc] = '/';

	// === Finding list ===

	char ListName [] = "00_OrderList.txt";
	for (i1 = 0x00, i2 = DirectoryLoc + 0x01, i0 = ' '; i0 != 0x00; )
	{
		i0 = ListName [i1++] & 0xFF;
		Directory [i2++] = i0;
	}
	FILE *List;
	if ((List = fopen (Directory, "rb")) == NULL)
	{
		printf ("    Error: Could not open \"%s\"\n", ListName);
	}
	else
	{
		// === Creating .ssf archieve ===

		for (i1 = DirectoryLoc - 0x01, i0 = 0x01; i0 != 0x0D && i0 != 0x0A && i0 != ' ' && i0 != 0x00; )
		{
			i0 = fgetc (List) & 0xFF;
			Directory [++i1] = i0;
		}
		Directory [i1] = 0x00;
		FILE *Output;
		if ((Output = fopen (Directory, "wb")) == NULL)
		{
			printf ("    Error: Could not create .ssf file\n");
		}
		else
		{
			// === Creating header ===

			char *Memory = NULL; int MemorySize = 0x00, MemoryLoc = 0x00;
			AllocChar (Memory, MemorySize, 0x1C);
			Memory [MemoryLoc++] = ' ';
			Memory [MemoryLoc++] = 'C';
			Memory [MemoryLoc++] = 'E';
			Memory [MemoryLoc++] = 'S';
			for (i7 = 0x00; i7 != 0x04; i7++)
			{
				MemoryLoc = ReadValue (List);
				i0 = ReadValue (List);
				Memory [MemoryLoc++] = i0;
				Memory [MemoryLoc++] = i0 >> 0x08;
				Memory [MemoryLoc++] = i0 >> 0x10;
				Memory [MemoryLoc++] = i0 >> 0x18;
			}

			// Finding file list

			i0 = 0x00, i1 = 0x00;
			while (i0 < '0' || i0 > '9' && i0 < 'A' || i0 > 'F')
			{
				i0 = fgetc (List) & 0xFF;
				if (i0 == ';')
				{
					do
					{
						i0 = fgetc (List) & 0xFF;
					}
					while (i0 != 0x0A);
				}
			}
			fseek (List, -0x01, SEEK_CUR);

			// === Counting files ===

			i7 = ftell (List);
			fseek (List, 0x00, SEEK_END);
			i6 = ftell (List);
			fseek (List, i7, SEEK_SET);
			for (FileCount = 0x00, StringSize = 0x00; i6 > ftell (List); FileCount++)
			{
				i1 = 0x00;
				do
				{
					i0 = fgetc (List) & 0xFF;
					if (i1 == 0x01)
					{
						if (i0 != '0' && i0 != ' ')
						{
							i1 = -0x40;
						}
					}
					else if (i1 == 0x02)
					{
						if (i0 != ' ')
						{
							StringSize++;
						}
					}
					if (i0 == ' ')
					{
						i1++;
					}
				}
				while (i0 != 0x0A);
				if (i1 > 0x00)
				{
					StringSize = (StringSize - 0x04) + 0x01;
				}
			}
			fseek (List, i7, SEEK_SET);
			MemoryLoc = 0x10;
			Memory [MemoryLoc++] = FileCount;
			Memory [MemoryLoc++] = FileCount >> 0x08;
			Memory [MemoryLoc++] = FileCount >> 0x10;
			Memory [MemoryLoc++] = FileCount >> 0x18;
			StringPos = 0x1C + (FileCount * 0x10);
			StringSize += StringPos;

			// === Getting pad size ===

			MemoryLoc = 0x08;
			PadMode = Memory [MemoryLoc++] & 0xFF;
			PadMode |= (Memory [MemoryLoc++] & 0xFF) << 0x08;
			PadMode |= (Memory [MemoryLoc++] & 0xFF) << 0x10;
			PadMode |= (Memory [MemoryLoc++] & 0xFF) << 0x18;
			if (PadMode == 0x00) // Offset 00000008 == 00000000
			{
				PadSize = 0x800;
			}
			else // Offset 00000008 == 03000000
			{
				PadSize = 0x80;
			}
			Offset = StringSize;

			// === Packing files ===

			printf ("    Packing files...\n\n");
			MemoryLoc = 0x1C;
			StringPos--;
			Directory [DirectoryLoc] = '/';
			for (i7 = 0x00; i7 < FileCount && Fail == 0x00; i7++)
			{
				i0 = Offset;
				if ((Offset & (PadSize - 0x01)) != 0x00)
				{
					Offset = (Offset & (-PadSize)) + PadSize;
				}
				AllocChar (Memory, MemorySize, Offset);
				PadData (i0, Offset, Memory, PadMode);
				Type = ReadValue (List);
				Memory [MemoryLoc++] = Type;
				Memory [MemoryLoc++] = Type >> 0x08;
				Memory [MemoryLoc++] = Type >> 0x10;
				Memory [MemoryLoc++] = Type >> 0x18;
				NameType = ReadValue (List);
				Memory [MemoryLoc++] = Offset;
				Memory [MemoryLoc++] = Offset >> 0x08;
				Memory [MemoryLoc++] = Offset >> 0x10;
				Memory [MemoryLoc++] = Offset >> 0x18;
				FileNameLoc = StringPos + 0x01;
				for (i1 = DirectoryLoc, i2 = 0x00, i0 = 0x00; i0 != ' '; )
				{
					i0 = fgetc (List) & 0xFF;
					Directory [++i1] = i0;
					if (NameType == 0x00)
					{
						if (i0 == '.')
						{
							i2 = StringPos + 0x01;
						}
						Memory [++StringPos] = i0;
					}
				}
				Directory [i1] = 0x00;
				if (i2 != 0x00)
				{
					StringPos = i2;
				}
				Memory [StringPos] = 0x00;
				printf ("    > %s\n", Directory);
				FILE *Input;
				if ((Input = fopen (Directory, "rb")) == NULL)
				{
					Fail = 0x01;
					printf ("      Error: Could not open file\n");
				}
				else
				{
					fseek (Input, 0x00, SEEK_END);
					Size = ftell (Input);
					rewind (Input);
					AllocChar (Memory, MemorySize, MemorySize + Size);
					for (i4 = 0x00; i4 < Size; i4++)
					{
						Memory [Offset++] = fgetc (Input);
					}
					fclose (Input);
					Memory [MemoryLoc++] = Size;
					Memory [MemoryLoc++] = Size >> 0x08;
					Memory [MemoryLoc++] = Size >> 0x10;
					Memory [MemoryLoc++] = Size >> 0x18;
					if (NameType == 0x00)
					{
						i0 = FileNameLoc - (0x1C + (FileCount * 0x10));
						Memory [MemoryLoc++] = i0;
						Memory [MemoryLoc++] = i0 >> 0x08;
						Memory [MemoryLoc++] = i0 >> 0x10;
						Memory [MemoryLoc++] = i0 >> 0x18;
					}
					else
					{
						Memory [MemoryLoc++] = NameType;
						Memory [MemoryLoc++] = NameType >> 0x08;
						Memory [MemoryLoc++] = NameType >> 0x10;
						Memory [MemoryLoc++] = NameType >> 0x18;
					}
				}
			}
			if ((Offset & (PadSize - 0x01)) != 0x00)
			{
				Offset = (Offset & (-PadSize)) + PadSize;
			}
			AllocChar (Memory, MemorySize, Offset);
			StringPos++;
			i0 = StringPos;
			if ((StringPos & (PadSize - 0x01)) != 0x00)
			{
				StringPos = (StringPos & (-PadSize)) + PadSize;
			}
			PadData (i0, StringPos, Memory, PadMode);
			MemoryLoc = 0x18;
			Memory [MemoryLoc++] = MemorySize;
			Memory [MemoryLoc++] = MemorySize >> 0x08;
			Memory [MemoryLoc++] = MemorySize >> 0x10;
			Memory [MemoryLoc++] = MemorySize >> 0x18;

			// === Finish ===

			fwrite (Memory, MemorySize, 0x01, Output);
			fclose (Output);
			free (Memory);
		}
		fclose (List);
	}
}

// =================================================================================================
// -------------------------------------------------------------------------------------------------
// Main routine
// -------------------------------------------------------------------------------------------------

int main (int ArgNumber, char **ArgList, char **EnvList)

{
	printf (".ssf ( CES) Compactor - by MarkeyJester\n");
	int ArgCount, i0, i1, i2, i3, i4, i5, i6, i7;
	if (ArgNumber <= 0x01)
	{
		printf ("\n -> Drag and drop a folder of files onto this application\n");
	}
	else
	{
		char Directory [0x0400];
		for (ArgCount = 0x01; ArgCount < ArgNumber; ArgCount++)
		{
			char *ArgListPointer = ArgList [ArgCount];
			for (i0 = ' ', i1 = 0x00; i0 != 0x00; )
			{
				i0 = ArgListPointer [i1] & 0xFF;
				Directory [i1++] = i0;
			}
			printf ("\n -> %s\n", Directory);
			CompactFile (Directory);
		}
	}
	printf ("\nPress enter key to exit...\n");
	fflush (stdin);
	getc (stdin);
	return (0x00);
}

// =================================================================================================
